package dev.peerat.parser.java;

import static org.junit.Assert.assertNull;
import static org.junit.jupiter.api.Assertions.assertEquals;

import java.util.List;
import java.util.function.Function;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.TestInstance.Lifecycle;

import dev.peerat.parser.Parser;
import dev.peerat.parser.java.tree.JavaTreeType;
import dev.peerat.parser.java.value.ArrayAccessValue;
import dev.peerat.parser.java.value.BiValue;
import dev.peerat.parser.java.value.LambdaValue;
import dev.peerat.parser.java.value.MethodCallValue;
import dev.peerat.parser.java.value.ModifierValue;
import dev.peerat.parser.java.value.StaticValue;
import dev.peerat.parser.java.value.TriValue;
import dev.peerat.parser.java.value.Value;
import dev.peerat.parser.java.value.Value.ValueContainer;
import dev.peerat.parser.java.value.VariableAccessValue;
import dev.peerat.parser.java.visitor.JavaVisitor;
import dev.peerat.parser.visitor.VisitorBag;

@TestInstance(Lifecycle.PER_CLASS)
public class ValueTypesTests {
	
	private Parser<JavaElement> parser;
	
	@BeforeAll
	public void init(){
		parser = new JavaParser(JavaTreeType.VALUE);
	}
	
	private Value parse(String value) throws Exception{
		Container container = new Container();
		parser.parse(value, container);
		return container.getValue();
	}
	
	private static class Container extends JavaElement implements ValueContainer{
		
		private Value value;
		
		@Override
		public void addValue(Value value) {
			this.value = value;
		}
		
		public Value getValue(){
			return this.value;
		}

		@Override
		public void build(Builder builder) throws Exception {}

		@Override
		public <E extends JavaElement> E find(Function<JavaElement, Boolean> finder){return null;}

		@Override
		public <E extends JavaElement> void findAll(Function<JavaElement, Boolean> finder, List<E> list){}

		@Override
		public VisitorBag visit(JavaVisitor<?> visitor){return null;}
		
	}
	
	private <T> T assertInstance(Object obj, java.lang.Class<? extends T> type){
		if(type.isAssignableFrom(obj.getClass())) return (T) obj;
		throw new IllegalArgumentException(obj.getClass()+" is not assignable to type "+type);
	}

	@Test
	public void onStaticValue() throws Exception{
		
		//ALL FORMATED VALUES LIKE DOUBLE, FLOAT, BINARY, CHARS(?),.....
		
		StaticValue v;
		
		v = assertInstance(parse("8"), StaticValue.class);
		assertEquals("8", v.getToken().getValue());		
		v = assertInstance(parse("value"), StaticValue.class);
		assertEquals("value", v.getToken().getValue());	
		v = assertInstance(parse("\"value\""), StaticValue.class);
		assertEquals("\"value\"", v.getToken().getValue());
		v = assertInstance(parse("'a'"), StaticValue.class);
		assertEquals("'a'", v.getToken().getValue());
	}
	
	@Test
	public void onBiValue() throws Exception{
		
		BiValue v;
		
		v = assertInstance(parse("8+8"), BiValue.class);
		assertEquals("8", assertInstance(v.left(), StaticValue.class).getToken().getValue());
		assertEquals("+", v.getAction());
		assertEquals("8", assertInstance(v.right(), StaticValue.class).getToken().getValue());
		
		v = assertInstance(parse("8+offset"), BiValue.class);
		assertEquals("8", assertInstance(v.left(), StaticValue.class).getToken().getValue());
		assertEquals("+", v.getAction());
		assertEquals("offset", assertInstance(v.right(), StaticValue.class).getToken().getValue());
		
		v = assertInstance(parse("8+offset > 8"), BiValue.class);
		BiValue right = assertInstance(v.right(), BiValue.class);
		assertEquals("8", assertInstance(v.left(), StaticValue.class).getToken().getValue());
		assertEquals("+", v.getAction());
		assertEquals("offset", assertInstance(right.left(), StaticValue.class).getToken().getValue());
		assertEquals(">", right.getAction());
		assertEquals("8", assertInstance(right.right(), StaticValue.class).getToken().getValue());
	}
	
	@Test
	public void onTriValue() throws Exception{
		
		TriValue v;
		
		v = assertInstance(parse("bool ? true : false"), TriValue.class);
		assertEquals("bool", assertInstance(v.getChecker(), StaticValue.class).getToken().getValue());
		assertEquals("true", assertInstance(v.success(), StaticValue.class).getToken().getValue());
		assertEquals("false", assertInstance(v.fail(), StaticValue.class).getToken().getValue());
		
		v = assertInstance(parse("bool ? 8+8 : false"), TriValue.class);
		assertEquals("bool", assertInstance(v.getChecker(), StaticValue.class).getToken().getValue());
		BiValue bv = assertInstance(v.success(), BiValue.class);
		assertEquals("8", assertInstance(bv.left(), StaticValue.class).getToken().getValue());
		assertEquals("+", bv.getAction());
		assertEquals("8", assertInstance(bv.right(), StaticValue.class).getToken().getValue());
		assertEquals("false", assertInstance(v.fail(), StaticValue.class).getToken().getValue());
		
		v = assertInstance(parse("bool ? rebool ? true : bool : false"), TriValue.class);
		assertEquals("bool", assertInstance(v.getChecker(), StaticValue.class).getToken().getValue());
		assertEquals("false", assertInstance(v.fail(), StaticValue.class).getToken().getValue());
		TriValue success = assertInstance(v.success(), TriValue.class);
		assertEquals("rebool", assertInstance(success.getChecker(), StaticValue.class).getToken().getValue());
		assertEquals("true", assertInstance(success.success(), StaticValue.class).getToken().getValue());
		assertEquals("bool", assertInstance(success.fail(), StaticValue.class).getToken().getValue());
		
		v = assertInstance(parse("(rebool ? true : bool) ? true : false"), TriValue.class);
		TriValue checker = assertInstance(v.getChecker(), TriValue.class);
		assertEquals("rebool", assertInstance(checker.getChecker(), StaticValue.class).getToken().getValue());
		assertEquals("true", assertInstance(checker.success(), StaticValue.class).getToken().getValue());
		assertEquals("bool", assertInstance(checker.fail(), StaticValue.class).getToken().getValue());
		assertEquals("true", assertInstance(v.success(), StaticValue.class).getToken().getValue());
		assertEquals("false", assertInstance(v.fail(), StaticValue.class).getToken().getValue());
	}
	
	@Test
	public void onModifierValue() throws Exception{
		
		ModifierValue v;
		v = assertInstance(parse("!a"), ModifierValue.class);
		assertEquals("!", v.getModifier().getValue());
		assertEquals("a", assertInstance(v.getValue(), StaticValue.class).getToken().getValue());
		
		v = assertInstance(parse("!(a || b)"), ModifierValue.class);
		assertEquals("!", v.getModifier().getValue());
		BiValue bv = assertInstance(v.getValue(), BiValue.class);
		assertEquals("a", assertInstance(bv.left(), StaticValue.class).getToken().getValue());
		assertEquals("||", bv.getAction());
		assertEquals("b", assertInstance(bv.right(), StaticValue.class).getToken().getValue());
	}
	
	@Test
	public void onMethodCallValue() throws Exception{
		MethodCallValue v;
		v = assertInstance(parse("test()"), MethodCallValue.class);
		assertNull(v.base());
		assertEquals("test", v.getToken().getValue());
		assertNull(v.getGeneric());
		assertEquals(0, v.getParameters().size());
		
		v = assertInstance(parse("Test.test()"), MethodCallValue.class);
		
		v = assertInstance(parse("JavaParserTest.test(8+8).equals(16)"), MethodCallValue.class);
		assertEquals("equals", v.getToken().getValue());
		assertEquals(1, v.getParameters().size());
		assertEquals("16", assertInstance(v.getParameters().get(0), StaticValue.class).getToken().getValue());
		v = assertInstance(v.base(), MethodCallValue.class);
		assertEquals(1, v.getParameters().size());
		BiValue bv = assertInstance(v.getParameters().get(0), BiValue.class);
		assertEquals("8", assertInstance(bv.left(), StaticValue.class).getToken().getValue());
		assertEquals("+", bv.getAction());
		assertEquals("8", assertInstance(bv.right(), StaticValue.class).getToken().getValue());
		assertEquals("test", v.getToken().getValue());
		assertEquals("JavaParserTest", assertInstance(v.base(), StaticValue.class).getToken().getValue());
		
		v = assertInstance(parse("main.test.here.localvar.run(6)[0].validate(7)"), MethodCallValue.class); //main.test.here.localvar.run(6)[0].validate(7)
		assertEquals("validate", v.getToken().getValue());
		assertEquals(1, v.getParameters().size());
		StaticValue param = assertInstance(v.getParameters().get(0), StaticValue.class);
		assertEquals("7", param.getToken().getValue());
		ArrayAccessValue array = assertInstance(v.base(), ArrayAccessValue.class); //main.test.here.localvar.run(6)[0]
		param = assertInstance(array.getAccessor(), StaticValue.class);
		assertEquals("0", param.getToken().getValue());
		v = assertInstance(array.base(), MethodCallValue.class); //main.test.here.localvar.run(6)
		assertEquals("run", v.getToken().getValue());
		param = assertInstance(v.getParameters().get(0), StaticValue.class);
		assertEquals("6", param.getToken().getValue());
		VariableAccessValue access = assertInstance(v.base(), VariableAccessValue.class);
		assertEquals("localvar", access.getVariable().getValue());
		access = assertInstance(access.base(), VariableAccessValue.class);
		assertEquals("here", access.getVariable().getValue());
		access = assertInstance(access.base(), VariableAccessValue.class);
		assertEquals("test", access.getVariable().getValue());
		assertEquals("main", assertInstance(access.base(), StaticValue.class).getToken().getValue());
	}
	
	@Test
	public void onLambdaValue() throws Exception{
		LambdaValue lv = assertInstance(parse("() -> {}"), LambdaValue.class);
		assertEquals(0, lv.getParameters().size()+lv.getOperations().size());
		
		lv = assertInstance(parse("(a) -> {}"), LambdaValue.class);
		assertEquals(1, lv.getParameters().size());
		assertNull(lv.getParameters().get(0).getType());
		assertEquals("a", lv.getParameters().get(0).getName().getValue());
		assertEquals(0, lv.getOperations().size());
		lv = assertInstance(parse("(a,b) -> {}"), LambdaValue.class);
		assertEquals(2, lv.getParameters().size());
		assertNull(lv.getParameters().get(0).getType());
		assertEquals("a", lv.getParameters().get(0).getName().getValue());
		assertNull(lv.getParameters().get(1).getType());
		assertEquals("b", lv.getParameters().get(1).getName().getValue());
		assertEquals(0, lv.getOperations().size());
		lv = assertInstance(parse("(int a,b) -> {}"), LambdaValue.class);
		assertEquals("int", lv.getParameters().get(0).getType().getValue());
		assertEquals("a", lv.getParameters().get(0).getName().getValue());
		assertNull(lv.getParameters().get(1).getType());
		assertEquals("b", lv.getParameters().get(1).getName().getValue());
		assertEquals(0, lv.getOperations().size());
		lv = assertInstance(parse("(int a, String b) -> {}"), LambdaValue.class);
		assertEquals("int", lv.getParameters().get(0).getType().getValue());
		assertEquals("a", lv.getParameters().get(0).getName().getValue());
		assertEquals("String", lv.getParameters().get(1).getType().getValue());
		assertEquals("b", lv.getParameters().get(1).getName().getValue());
		assertEquals(0, lv.getOperations().size());
		lv = assertInstance(parse("(int a, String b) -> true"), LambdaValue.class);
		assertEquals("int", lv.getParameters().get(0).getType().getValue());
		assertEquals("a", lv.getParameters().get(0).getName().getValue());
		assertEquals("String", lv.getParameters().get(1).getType().getValue());
		assertEquals("b", lv.getParameters().get(1).getName().getValue());
		assertEquals(1, lv.getOperations().size());
		assertEquals("true", assertInstance(lv.getOperations().get(0), StaticValue.class).getToken().getValue());
		lv = assertInstance(parse("(int a, String b) -> { test(); }"), LambdaValue.class);
		assertEquals("int", lv.getParameters().get(0).getType().getValue());
		assertEquals("a", lv.getParameters().get(0).getName().getValue());
		assertEquals("String", lv.getParameters().get(1).getType().getValue());
		assertEquals("b", lv.getParameters().get(1).getName().getValue());
		assertEquals(1, lv.getOperations().size());
		assertEquals("test", assertInstance(lv.getOperations().get(0), MethodCallValue.class).getToken().getValue());
		lv = assertInstance(parse("(int a, String b) -> test()"), LambdaValue.class);
		assertEquals("int", lv.getParameters().get(0).getType().getValue());
		assertEquals("a", lv.getParameters().get(0).getName().getValue());
		assertEquals("String", lv.getParameters().get(1).getType().getValue());
		assertEquals("b", lv.getParameters().get(1).getName().getValue());
		assertEquals(1, lv.getOperations().size());
		assertEquals("test", assertInstance(lv.getOperations().get(0), MethodCallValue.class).getToken().getValue());
		
		lv = assertInstance(parse("(int a, String b) -> {test(); testu();}"), LambdaValue.class);
		assertEquals("int", lv.getParameters().get(0).getType().getValue());
		assertEquals("a", lv.getParameters().get(0).getName().getValue());
		assertEquals("String", lv.getParameters().get(1).getType().getValue());
		assertEquals("b", lv.getParameters().get(1).getName().getValue());
		assertEquals(2, lv.getOperations().size());
		assertEquals("test", assertInstance(lv.getOperations().get(0), MethodCallValue.class).getToken().getValue());
		assertEquals("testu", assertInstance(lv.getOperations().get(1), MethodCallValue.class).getToken().getValue());
		
		assertInstance(parse("(hello)"), StaticValue.class);
		assertInstance(parse("(8+8)"), BiValue.class);
		
	}
}
